import logging.config

__all__ = ['set_log_conf']

_LOG_CONFIG = {
    'version': 1,
    'disable_existing_loggers': True,
    'formatters': {
        'standard': {
            'format': '%(levelname)s %(asctime)s %(message)s',
            'datefmt': "%H:%M:%S",
        },
        'debug': {
            'format': '[%(levelname)s] %(module)s:%(funcName)s %(name)s: %(message)s',
            'datefmt': "%H:%M:%S",
        },
    },
    'handlers': {
        'standard': {
            'level': 'WARNING',
            'formatter': 'standard',
            'class': 'logging.StreamHandler',
        },
        'debug': {
            'level': 'DEBUG',
            'formatter': 'debug',
            'class': 'logging.StreamHandler',
        },
    },
    'loggers': {
        '': {},
    },
}

_STANDARD_LOGGER = {
    'handlers': ['standard'],
    'level': 'INFO',
    'propagate': False,
}

_DEBUG_LOGGER = {
    'handlers': ['standard'],
    'level': 'INFO',
    'propagate': False,
}


def set_log_conf(debug: bool = False):
    if debug:
        _LOG_CONFIG['loggers'][''] = _DEBUG_LOGGER
    else:
        _LOG_CONFIG['loggers'][''] = _STANDARD_LOGGER
    logging.config.dictConfig(_LOG_CONFIG)



try:
    from IPython.display import Code
except ImportError:
    Code = None

def jupyter_code(func_code: str):
    global Code

    if not Code:
        try:
            from IPython.display import Code
        except ImportError:
            raise ImportError('Could not import IPython.display.Code')
    return Code(func_code)


__all__ = ['jupyter_code']



from pathlib import Path
from datetime import datetime as dt
from functools import wraps
from typing import List


from ..file import save_plt_file, read_plt_file, PltFile


class PltProject(object):
    def __init__(self, folder: Path or str, datefmt: str = '%d-%b-%H-%M-%S', format: str = 'plt'):
        self.folder = Path(folder)
        self.folder.mkdir(exist_ok=True)
        self.datefmt = datefmt
        self.format = format

    def save(self, func,
             args: tuple = (),
             kwargs: dict = None,
             func_name: str = None,
             rewrite: bool = False) -> Path:
        func_name = func_name or func.__name__
        file_path = self.folder / _func_name(func_name, self.datefmt, rewrite, self.format)
        save_plt_file(file_path, func, *args, **(kwargs or {}))
        return file_path

    def auto_save(self, name: str = None, rewrite: bool = False):
        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                func(*args, **kwargs)
                self.save(func, args=args, kwargs=kwargs, func_name=name, rewrite=rewrite)

            return wrapper
        return decorator

    def list_files(self) -> List[str]:
        return [path.stem for path in self.folder.glob(f'*.{self.format}')]

    def load_file(self, name: str) -> PltFile:
        return read_plt_file(self.folder / f'{name}.{self.format}')

    def __getitem__(self, name: str):
        return self.load_file(name)

    @property
    def time_str(self):
        return time_str(self.datefmt)


def _func_name(func_name: str, datefmt: str, rewrite: bool, format: str = 'plt'):
    if rewrite:
        return f'{func_name}.{format}'
    else:
        return f'{func_name}-{time_str(datefmt)}.{format}'


def time_str(datefmt: str):
    return dt.now().strftime(datefmt)



import importlib
from types import ModuleType


__all__ = ['check_version']


try:
    from importlib.metadata import version  # python 3.8 or above

    def check_version(module: str or ModuleType) -> str or None:
        if isinstance(module, ModuleType):
            if module.__package__:
                try:
                    return version(module.__package__)
                except ImportError:
                    return check_version(module.__name__)
        else:
            try:
                return version(module)
            except ImportError:
                try:
                    return version(module.split('.')[0])
                except ImportError:
                    return


except ImportError:

    from packaging.version import parse as parse_version
    from packaging.version import Version

    def check_version(module: str or ModuleType) -> str or None:
        if isinstance(module, str):
            module = _str_to_module(module)

            if not module:
                return

        if hasattr(module, '__version__'):
            version = module.__version__
            if _version_is_valid(version):
                return version
        if hasattr(module, '__package__') and module.__package__:
            return check_version(module.__package__)


    def _version_is_valid(version_str: str) -> bool:
        return isinstance(parse_version(version_str), Version)


    def _str_to_module(module: str) -> ModuleType or None:
        try:
            module = globals()[module]
        except KeyError:
            try:
                return importlib.import_module(module)
            except ImportError:
                return



import sys
from datetime import datetime as dt


def get_metadata() -> dict:
    platform = sys.platform
    py_version = sys.version
    time = dt.utcnow().strftime('%d %B %Y %H:%M:%S')
    return dict(time=time, py_version=py_version, platform=platform)


if __name__ == '__main__':
    print(get_metadata())



import ast
import builtins
from typing import List, Set, Dict

__all__ = ['CheckUsedGlobals']


class CheckUsedGlobals(ast.NodeVisitor):
    builtins = dir(builtins)

    def __init__(self):
        self.modules: Set[str] = set()
        self.imported: Set[str] = set()
        self.args: Set[str] = set()
        self.assigned_names: Set[str] = set()
        self.used_globals: Set[str] = set()

    def _check_assigned(self, name):
        return (
                name in self.assigned_names or
                name in self.builtins or
                name in self.args
        )

    def clear(self):
        self.modules.clear()
        self.imported.clear()
        self.args.clear()
        self.assigned_names.clear()
        self.used_globals.clear()

    def visit_Import(self, node):

        for import_name in node.names:
            self.modules.add(import_name.name.split('.')[0])
            self.imported.add(import_name.name)

            if import_name.asname:
                self.assigned_names.add(import_name.asname)
            else:
                name = import_name.name.split('.')[-1]
                self.assigned_names.add(name)

        self.generic_visit(node)

    def visit_ImportFrom(self, node):
        self.modules.add(node.module.split('.')[0])

        for import_name in node.names:
            self.imported.add('.'.join([node.module, import_name.name]))
            if import_name.asname:
                self.assigned_names.add(import_name.asname)
            else:
                self.assigned_names.add(import_name.name)

        self.generic_visit(node)

    def visit_Assign(self, node):
        assigned = []
        for node in node.targets:
            if isinstance(node, ast.Name):
                assigned.append(node.id)
            elif isinstance(node, ast.Tuple):
                assigned += [n.id for n in node.elts]
        self.assigned_names.update(assigned)
        self.generic_visit(node)

    def visit_FunctionDef(self, node):
        self.assigned_names.add(node.name)
        self.args.update(
            _get_funcdef_arg_names(node, 'posonlyargs') +
            _get_funcdef_arg_names(node, 'args') +
            _get_funcdef_arg_names(node, 'kwonlyargs')
        )
        self.generic_visit(node)

    def visit_Global(self, node):
        self.used_globals.update(node.names)
        self.generic_visit(node)

    def visit_Name(self, node):
        if not self._check_assigned(node.id):
            self.used_globals.add(node.id)
        self.generic_visit(node)


def _get_funcdef_arg_names(node, attr) -> List[str]:
    arg_list = getattr(node.args, attr, None)
    if not arg_list:
        return []
    return list(map(lambda x: x.arg, getattr(node.args, attr)))



from typing import Union
from io import BytesIO
from contextlib import contextmanager
from pathlib import Path

import matplotlib.pyplot as plt


def get_diff_rc_params():
    return {
        k: v for k, v in plt.rcParams.items()
        if (
                v != plt.rcParamsDefault[k] and
                k != 'axes.prop_cycle'
        )
    }


class NotAPltFileError(TypeError):
    pass


class CorruptedPltFileError(IOError):
    pass


class NonSerializableDataError(TypeError):
    pass


@contextmanager
def _open_file_from_args(file: Union[str, Path, BytesIO], mode: str = 'rb', *args, **kwargs):
    if isinstance(file, Path):
        file = str(file.absolute())

    if isinstance(file, str):
        with open(file, mode, *args, **kwargs) as f:
            yield f
        return
    elif hasattr(file, 'read'):
        yield file  # we do not close it since we did not open it ourselves
        return
    raise TypeError(f'Expected a file or a path-like object, received {file} instead.')



def create_new_jupyter_cell(func_code: str):
    try:
        from IPython.core.getipython import get_ipython
    except ImportError:
        raise ImportError('Could not import IPython.core.getipython.')
    shell = get_ipython()
    payload = dict(
        source='set_next_input',
        text=func_code,
        replace=False,
    )
    shell.payload_manager.write_payload(payload, single=False)



import ast
from inspect import getsource
import logging
import importlib

from types import (
    FunctionType,
    ModuleType,
)

from typing import (
    NamedTuple,
    Dict,
    Any,
    List,
    Tuple,
)

from .func_dict import FuncDict
from .check_used_globals import CheckUsedGlobals
from .check_version import check_version
from ..utils import is_serializable, find_non_serializable

__all__ = ['FuncParser', 'parse_function']



def parse_function(fun: FunctionType) -> FuncDict:
    return FuncParser().parse_func(fun)


class FuncParser(object):
    def __init__(self):
        self.log = logging.getLogger(__name__)
        self.check_globals = CheckUsedGlobals()
        self.used_modules = set()
        self.import_lines = set()
        self.global_vars = {}
        self._unparsed_functions = {}
        self.parsed_functions = {}
        self.module_versions = {}
        self.func_dict = {}
        self.src_module: str = None
        self.entry_func: str = None

    def clear(self):
        self.used_modules.clear()
        self.import_lines.clear()
        self.global_vars.clear()
        self._unparsed_functions.clear()
        self.parsed_functions.clear()
        self.module_versions.clear()
        self.func_dict.clear()
        self.src_module = None
        self.entry_func = None

    def parse_func(self, fun: FunctionType) -> FuncDict:
        if not isinstance(fun, FunctionType):
            raise TypeError(f'Only FunctionType objects supported, got {type(fun).__name__}.')

        self.clear()

        self.entry_func = fun.__name__

        self.src_module = fun.__module__ if hasattr(fun, '__module__') else '__main__'

        self._parse_func(fun)
        self._update_module_versions(fun)
        return self.to_func_dict()

    def _parse_func(self, fun: FunctionType):
        if fun.__name__ in self.parsed_functions:
            return

        self.check_globals.clear()
        fun_code = getsource(fun)
        fun_code = _fix_func_code(fun_code)
        tree = ast.parse(fun_code)
        self.check_globals.visit(tree)

        self.used_modules.update(self.check_globals.modules)
        self._parse_globals(fun)
        self.parsed_functions[fun.__name__] = fun_code

        unparsed_functions = dict(self._unparsed_functions)
        self._unparsed_functions.clear()

        for fun in unparsed_functions.values():
            self._parse_func(fun)

    def _update_module_versions(self, fun: FunctionType):
        for module_name in self.used_modules:
            try:
                module = importlib.import_module(module_name)
                version = check_version(module)
            except (TypeError, ImportError):
                continue
            if version:
                self.module_versions[module_name] = version
            else:
                self.log.warning(f'Could not retrieve the version of the module {module_name}.')

    def _parse_globals(self, fun: FunctionType):
        for name in self.check_globals.used_globals:
            self._parse_global_name(fun, name)

    def _parse_global_name(self, fun, name) -> None:
        var, status = _get_global_var_from_name(name, fun)

        if not status:
            self.log.warning(f'Could not find a global variable {name}.')
            return

        if self._is_duplicate(var, name):
            return

        if self._parse_module_type(var, name):
            return

        if self._parse_imported_object(var, name):
            return

        if isinstance(var, FunctionType):
            self._unparsed_functions[name] = var
        elif is_serializable(var):
            self.global_vars[name] = var
        else:
            raise TypeError(f'Global var {name} is not serializable: {find_non_serializable(var)}.')

    def _parse_imported_object(self, var, name) -> bool:
        if hasattr(var, '__module__'):
            module_name = var.__module__
            if module_name != self.src_module:
                var_name = var.__name__ if hasattr(var, '__name__') else name
                self.import_lines.add(_get_import_line(var_name, module_name, name))
                self.used_modules.add(module_name.split('.')[0])
                return True
        return False

    def _is_duplicate(self, var, name) -> bool:
        if name in self._unparsed_functions and self._unparsed_functions[name] == var:
            return True
        if name in self.global_vars and self.global_vars[name] == var:
            return True
        return False

    def _parse_module_type(self, var: ModuleType, assigned_name) -> bool:
        if not isinstance(var, ModuleType):
            return False
        try:
            module_name = var.__name__
        except AttributeError:
            self.log.warning(f'Failed getting a module name: {var}, use {assigned_name} instead.')
            module_name, assigned_name = assigned_name, None
        self.import_lines.add(_get_import_line(module_name, assigned_name=assigned_name))
        self.used_modules.add(module_name.split('.')[0])
        return True

    def to_func_dict(self) -> FuncDict:
        return FuncDict(
            entry_func=self.entry_func,
            functions=dict(self.parsed_functions),
            modules=tuple(self.used_modules),
            import_lines=tuple(self.import_lines),
            module_versions=dict(self.module_versions),
            global_vars=dict(self.global_vars),
        )


def _get_global_var_from_name(name, fun) -> Tuple[Any, bool]:
    try:
        var = fun.__globals__[name]
    except KeyError:
        if name in globals():
            var = globals()[name]
        else:
            return None, False
    return var, True


def _get_import_line(name: str, from_str: str = None, assigned_name: str = None) -> str:
    if from_str:
        line = f'from {from_str} import {name}'
    else:
        line = f'import {name}'
    if assigned_name and assigned_name != name:
        line = f'{line} as {assigned_name}'
    return line


def _fix_func_code(func_code: str):
    func_code = _fix_indents(func_code, ' ')
    func_code = _fix_indents(func_code, '\t')
    func_code = _remove_decorators(func_code)

    return func_code


def _remove_decorators(func_code: str):
    counts = 0
    lines = func_code.split('\n')

    for line in lines:
        if line.startswith('def '):
            break
        counts += 1
    return '\n'.join(lines[counts:])


def _fix_indents(func_code: str, char: str):
    counts: int = 0
    lines = func_code.split('\n')

    for ch in lines[0].split(char):
        if ch:
            break
        counts += 1
    return '\n'.join([line[counts:] for line in lines])



from types import FunctionType
from typing import Union
from io import BytesIO
from pathlib import Path

from ..jupyter_support import create_new_jupyter_cell, jupyter_code

from ..parse import FuncDict
from .save_read import save_file, read_file
from .file_dict import get_file_dict

__all__ = ['PltFile', 'save_plt_file', 'read_plt_file']


class PltFile(object):
    def __init__(self, func: FunctionType = None, *args, _file_dict: FuncDict = None, **kwargs):
        self.file_dict: FuncDict
        if _file_dict:
            self.file_dict = _file_dict
        else:
            self.file_dict = get_file_dict(func, *args, **kwargs)

    def get_code_str(self, with_imports: bool = True, with_subfunctions: bool = True, with_globals: bool = True):
        code = ''

        entry_func_name = self.file_dict.entry_func
        entry_func = self.file_dict.functions[entry_func_name]
        sub_funcs = [func for func_name, func in self.file_dict.functions.items()
                     if func_name != entry_func_name]

        if with_imports:
            import_lines = '\n'.join(self.file_dict.import_lines)
            code += import_lines
            code += '\n\n'

        if with_globals:
            global_lines = '\n'.join([f'{k} = {v}' for k, v in self.file_dict.global_vars.items()])
            code += global_lines
            code += '\n\n'

        code += entry_func

        if with_subfunctions:
            code += '\n\n'
            code += '\n\n'.join(sub_funcs)

        return code

    @property
    def used_modules(self):
        return self.file_dict['modules']

    @classmethod
    def from_file(cls, file: Union[str, Path, BytesIO]):
        return cls(_file_dict=read_file(file))

    def save(self, file: Union[str, Path, BytesIO]):
        save_file(file, self.file_dict)

    def show_code(self, with_imports: bool = True, with_subfunctions: bool = True, with_globals: bool = True):
        code = self.get_code_str(with_imports, with_subfunctions, with_globals)
        try:
            create_new_jupyter_cell(code)
        except ImportError:
            return code

    def _ipython_display_(self):
        return jupyter_code(self.get_code_str())


def save_plt_file(file: Union[str, Path, BytesIO], func: FunctionType, *args, **kwargs) -> PltFile:
    plt_file = PltFile(func, *args, **kwargs)
    plt_file.save(file)
    return plt_file


def read_plt_file(file: Union[str, Path, BytesIO]) -> PltFile:
    return PltFile(_file_dict=FuncDict(**read_file(file)))



from setuptools import setup, find_packages
from pathlib import Path


def read(filename: str):
    with open(Path(__file__).parent / filename, mode='r', encoding='utf-8') as f:
        return f.read()


def get_version() -> str:
    local_dict = {}
    with open(str(Path(__file__).parent / 'src' / 'plt_pack' / '__version.py'), 'r') as f:
        exec(f.read(), {}, local_dict)

    return local_dict['__version__']


__version__ = get_version()

install_requires = [
    'numpy',
    'msgpack>=1.0.0',
    'msgpack-numpy>=0.4.7.1',
    'matplotlib',
    'packaging',
]

classifiers = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
]

setup(
    name='plt_pack',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    version=__version__,
    author='Vladimir Starostin',
    author_email='vladimir.starostin@uni-tuebingen.de',
    url='https://github.com/StarostinV/plt-pack',
    description='A packaging tool for storing and exchanging data & code bound in a single file.',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    license='MIT',
    python_requires='>=3.7.2',
    install_requires=install_requires,
    classifiers=classifiers,
)



from .msgpack_import import msgpack


__all__ = ['find_non_serializable', 'is_serializable']


def is_serializable(data) -> bool:
    try:
        msgpack.packb(data)
        return True
    except TypeError:
        return False


def find_non_serializable(data: dict or list):
    return _find_non_serializable(data)


def _find_non_serializable(data: dict or list, *, prev_keys=None) -> list:
    _found_keys = []

    prev_keys = prev_keys or []

    if isinstance(data, (list, tuple)):
        data = {str(i): v for i, v in enumerate(data)}

    for key, value in data.items():
        if isinstance(value, (list, tuple, dict)):
            _found_keys += _find_non_serializable(value, prev_keys=prev_keys + [key])
        else:
            if not is_serializable(value):
                _found_keys.append(prev_keys + [key])
    return _found_keys
